/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import java.awt.Image;
import java.awt.image.PixelGrabber;

public class DistributedShader extends Shader
{	
	protected float base;
	protected float specular;
	protected float reflective;
	protected float refraction;
	
	private void sumAmounts()
	{
		base = baseAmount;
		specular = base + specularAmount;
		reflective = specular + reflectiveAmount;
		refraction = reflective + refractionAmount;
	}
		
	/* ====================================================== */
	
	public Color getIntersectionColor( Ray ray, float t )
	{
		final Vector intersection = ray.getVector(t);
		final Vector normal = getSurfaceNormal(intersection);
		
		final float random = (float)(Math.random()) * refraction;
		
		if ( random < base )
		{
			final Color surfaceColor = getSurfaceColor( intersection );
			
			return getDiffuseColor( intersection, normal ).filter( surfaceColor );
		}
		else if ( random < specular )
		{
			final Vector reflection = ray.direction.sub(normal.scale(2.0f*(normal.dot(ray.direction)))).normalized();
			
			return getSpecularColor( intersection, reflection, normal );
		}
		else if ( random < reflective )
		{
			final Vector reflection = ray.direction.sub(normal.scale(2.0f*(normal.dot(ray.direction)))).normalized();
			
			return getReflection( intersection, reflection, ray.refraction );
		}
		else // refraction
		{
			final Color surfaceColor = getSurfaceColor( intersection );
			
			return getTransmission( intersection, ray.direction, normal, ray.refraction, surfaceColor );
		}
	}
	
	/* ====================================================== */
	
	protected Color getDiffuseColor( Vector intersection, Vector normal )
	{
		Color diffuse = Color.BLACK;
		
		Light[] lights = scene.getLights();

		nextlight:
		for (int i = 0; i < lights.length; i++)
		{
			// Get the vector from the intersection point to the current light
			Vector diff = lights[i].position.sub(intersection);
			
			// Get the length of the above light-vector
			float length = diff.length();

			// Turn the light-vector into a unit vector
			Vector lightDir = diff.normalized();

			float dot = normal.dot(lightDir);

			if (dot <= 0)
			{
				// Light is behind the surface
				continue;
			}
			
			/*
			 * Search for objects blocking the light
			 */

			Ray lightRay = new Ray(intersection, lightDir, 1, 1);

			AbstractObject[] objs = scene.getObjects();
			
			// Loop through every object on the path of our ray
			for ( int j = 0; j < objs.length; j++ )
			{
				// Get the intersection distance
				final float s = objs[j].getNearestIntersection(lightRay).t;
				
				// If the distance is greater than our threshold but
				// less than the distance to the light, then the object
				// must lie between the light and the intersection point.
				// Thus we've determinded that there's an object covering
				// the light, and don't need to add the light.
				if ( s > 0.001 && s < length )
					continue nextlight;
			}
			
			Color lightColor = lights[i].color.attenuate(length);
			
			/*
			 * Diffuse lighting
			 */
			 diffuse = diffuse.add( lightColor.scale(dot) );
		}
		
		return diffuse;
	}
	
	protected Color getSpecularColor( Vector intersection, Vector reflectDir, Vector normal )
	{
		Color specular = Color.BLACK;
		
		Light[] lights = scene.getLights();

		nextlight:
		for (int i = 0; i < lights.length; i++)
		{
			// Get the vector from the intersection point to the current light
			Vector diff = lights[i].position.sub(intersection);
			
			// Get the length of the above light-vector
			float length = diff.length();

			// Turn the light-vector into a unit vector
			Vector lightDir = diff.normalized();

			float dot = normal.dot(lightDir);

			if (dot <= 0)
			{
				// Light is behind the surface
				continue;
			}
			
			/*
			 * Search for objects blocking the light
			 */

			Ray lightRay = new Ray(intersection, lightDir, 1, 1);

			AbstractObject[] objs = scene.getObjects();
			
			// Loop through every object on the path of our ray
			for ( int j = 0; j < objs.length; j++ )
			{
				// Get the intersection distance
				final float s = objs[j].getNearestIntersection(lightRay).t;
				
				// If the distance is greater than our threshold but
				// less than the distance to the light, then the object
				// must lie between the light and the intersection point.
				// Thus we've determinded that there's an object covering
				// the light, and don't need to add the light.
				if ( s > 0.001 && s < length )
					continue nextlight;
			}
			
			Color lightColor = lights[i].color.attenuate(length);

			/*
			 * Specular lighting
			 */
			float reflectDot = reflectDir.dot( lightDir );
			
			if ( reflectDot > 0 )
			{
				float specularIllumination = (float)Math.pow( reflectDot, specularExponent );
			
				specular = specular.add( lightColor.scale( specularIllumination * specularAmount ) );
			}
		}
		
		return specular;
	}
	
	public void setColor( Color color, float amount )
	{
		super.setColor( color, amount );
		sumAmounts();
	}
	
	public void setSpecular( float exponent, float amount )
	{
		super.setSpecular( exponent, amount );
		sumAmounts();
	}
	
	public void setTranslucency( float refract, float amount )
	{
		super.setTranslucency( refract, amount );
		sumAmounts();
	}
	
	public void setReflectivity( float amount )
	{
		super.setReflectivity( amount );
		sumAmounts();
	}
	
	/* ====================================================== */
	
	public DistributedShader( Scene scene )
	{
		super( scene );
	}
}
